"""
quickson but for the masterclass users; using regex.

This **can** get out of hand.
"""

import re

SYNTAX_SEPARATOR_REGEX = re.compile(r"(?:[^/]|\\/)+")


def iter_regexson_matches(data, *regexsons):

    if not regexsons:
        yield data
        return

    if not isinstance(data, (list, dict)):
        raise TypeError("Data must be a dict or list")

    worker, *rest = regexsons

    if isinstance(data, list):

        if not worker.isdigit():
            raise ValueError("List index must be an integer")

        yield from iter_regexson_matches(data[int(worker)], *rest)
    else:
        for (key, value) in data.items():
            regexp = re.compile(worker)

            if regexp.match(key):
                yield from iter_regexson_matches(value, *rest)


def iter_regexson_matches_from_string(regexson_string, data):
    """
    Iterate over the matches of a regexson string.
    """
    yield from iter_regexson_matches(
        data, *SYNTAX_SEPARATOR_REGEX.findall(regexson_string)
    )



import asyncio
import contextvars
import inspect


def to_asyncio(f):
    """
    Convert a blocking function to a proper asyncio coroutine.
    """

    async def __inner__(*args, **kwargs):

        loop = asyncio.get_running_loop()
        ctx = contextvars.copy_context()

        return await loop.run_in_executor(None, ctx.run, f, *args, **kwargs)

    return __inner__


class AsyncClassFactory(type):
    """
    Sets up your class' functions with their
    asynchronous counterparts.
    """

    def __new__(cls, qualified_name, bases, namespace):

        target_class = super().__new__(cls, qualified_name, bases, namespace)
        class_dir = dir(target_class)

        for attrib in class_dir:

            item = getattr(target_class, attrib, None)

            if (
                item is None
                or (not inspect.isfunction(item))
                or attrib.startswith(("_", "async_"))
                or f"async_{attrib}" in class_dir
                or asyncio.iscoroutinefunction(item)
            ):
                continue

            setattr(target_class, f"async_{attrib}", to_asyncio(item))

        return target_class



import base64
import textwrap

BASE64_TABLE = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"


def cb64decode(data, *, table=BASE64_TABLE):
    return "".join(
        map(
            chr,
            base64.b64decode(
                (data + "=" * (len(data) % 4))
                .translate(str.maketrans(table, BASE64_TABLE))
                .encode()
            ),
        )
    )


def cb64encode(data, *, table=BASE64_TABLE):
    return "".join(
        table[int(segment.ljust(6, "0"), 2) if len(segment) < 6 else int(segment, 2)]
        for segment in textwrap.wrap(
            "".join(bin(ord(c)).lstrip("0b").rjust(8, "0") for c in data), 6
        )
    )



import re


URL_HOST_REGEX = re.compile(
    r"^(?P<scheme_prefix>(?P<scheme>.*?):.*?//)?(?P<basic_authorization>[a-zA-Z0-9-]+?:[a-zA-Z0-9-]+?@)?(?P<host>(?:[a-zA-Z0-9-]+?\.)+[a-zA-Z0-9-]+(?P<port>:\d+)?)(?P<path>(?:/[^/?&#]*)*)?"
)


def _get_segment(segment, retain, replacement_segment, force_presense, escape_segment):

    if retain and segment:
        if escape_segment:
            regex_segment = f"(?:{re.escape(segment)})"
        else:
            regex_segment = f"(?:{segment})"
    else:
        regex_segment = replacement_segment

    if not force_presense:
        regex_segment += "?"

    return regex_segment


def regexify_url(
    url: str,
    flags=re.IGNORECASE,
    *,
    extra: str = "",
    extra_re: str = "",
    match_subdomains: bool = True,
    retain_scheme: bool = True,
    retain_basic_authorization: bool = True,
    retain_path: bool = True,
    force_scheme_presence: bool = False,
    force_basic_authorization_presence: bool = False,
    force_path_presence: bool = False,
    strict: bool = False,
) -> re.Pattern:
    """
    Automatically convert a URL to a regex pattern.


    :param url: The URL to convert.
    :param flags: The regex flags to use.
    :param extra: Extra non-regex pattern to match.
    :param extra_re: Extra regex pattern to match.
    :param match_subdomains: Whether to match subdomains.
    :param retain_scheme: Whether to retain the scheme.
    :param retain_basic_authorization: Whether to retain the basic authorization.
    :param retain_path: Whether to retain the path.
    :param force_scheme_presence: Whether to force the scheme to be present.
    :param force_basic_authorization_presence: Whether to force the basic authorization to be present.
    :param force_path_presence: Whether to force the path to be present.
    :param strict: Whether to use strict mode.
    :return: The regex pattern.
    """

    regex_match = URL_HOST_REGEX.match(url)

    if not regex_match:
        raise ValueError(f"Could not resolve URL from {url!r}")

    url_segments = regex_match.groupdict()

    host_regex_segment = re.escape(url_segments["host"])

    if match_subdomains:
        host_regex_segment = r"(?:[a-zA-Z0-9-]+?\.)*" + host_regex_segment

    if url_segments["port"]:
        host_regex_segment += url_segments["port"]
    else:
        host_regex_segment += "(?::\d+)?"

    full_re = (
        _get_segment(
            url_segments.get("scheme_prefix"),
            retain_scheme,
            r"(?:.*?:.*?//)",
            force_scheme_presence,
            True,
        )
        + _get_segment(
            url_segments.get("basic_authorization"),
            retain_basic_authorization,
            r"(?:[a-zA-Z0-9-]+?:[a-zA-Z0-9-]+?@)",
            force_basic_authorization_presence,
            True,
        )
        + host_regex_segment
        + _get_segment(
            url_segments.get("path"),
            retain_path,
            "",
            force_path_presence,
            True,
        )
        + re.escape(extra)
        + extra_re
    )

    if strict:
        full_re = "^{}$".format(full_re)

    return re.compile(full_re, flags=flags)


if __name__ == "__main__":

    url = "abc.xyz/abc/"

    regex = regexify_url(url)

    assert regex.fullmatch("abc.xyz")
    assert regex.fullmatch("abc.xyz/abc/")
    assert regex.fullmatch("https://abc.xyz/abc/")
    assert regex.fullmatch("https://username:password@abc.xyz/abc/")
    assert regex.fullmatch("https://username:password@abc.xyz:8080/abc/")



from .cryptography import custom_b64
from .paginator import iter_contentaware_segments
from .parsing import objectnotation, ranges
from .regexp import regexify_url
from .search import iter_regex_results



import re


def iter_conditions(string, media_matchers=r"[0-9.]", media_separators=r"[:=-]"):
    """
    Iterate over the range conditions in a string.
    """
    if not string:
        yield lambda _: True

    regexp = re.compile(
        rf"(?:(?P<start>{media_matchers}*)(?P<separator>{media_separators}))?(?P<end>{media_matchers}+)"
    )

    for match in regexp.finditer(string):

        start, end = (float(_) if _ else None for _ in match.group("start", "end"))

        contains_separator = match.group("separator") is not None

        if not start:
            if contains_separator:
                yield lambda value, end=end: value <= end
            else:
                yield lambda value, end=end: value == end

        else:
            if start > end:
                start, end = end, start

            yield lambda value, start=start, end=end: start <= value <= end


def match_conditions(
    value,
    condition_string,
    selection_func=any,
    *,
    media_matchers=r"[0-9.]",
    media_separators=r"[:=-]",
):
    """
    Match a value against a list of conditions.
    """
    return selection_func(
        condition(value)
        for condition in iter_conditions(
            condition_string, media_matchers, media_separators
        )
    )



def iter_contentaware_segments(
    string: str, max_length: int, *, separators=("\n", "\t", "\x10")
):
    """
    Iterate over string segments, splitting at separator characters, at their maximum length.
    """

    while string:
        separation_index = max(
            string.rfind(separator, 0, max_length) for separator in separators
        )

        if separation_index == -1:
            separation_index = max_length

        yield string[:separation_index]
        string = string[separation_index:].lstrip("".join(separators))



from typing import Callable, Iterable, TypeVar

import re

search_type = TypeVar("search_type")


def iter_regex_results(
    query: "str",
    possibilities: "Iterable[search_type]",
    *,
    processor: "Callable[[search_type], str] | None" = None
):
    """
    Use regex to search for a string in a list of possibilities.

    :param query: The query to search for.
    :param possibilities: The list of possibilities to search in.
    :param processor: A function to process each possibility.

    :yield: The matching possibilities, level of match.
    """
    pattern = re.compile(
        r"(.*?)".join(map(re.escape, query.strip())) + r"(.*)", flags=re.IGNORECASE
    )

    for raw_search_value in possibilities:

        if processor is not None:
            search_value = processor(raw_search_value)
        else:
            search_value = raw_search_value

        match = pattern.match(search_value)

        if match:
            yield search_value, (1 - sum(map(len, match.groups())) / len(search_value))



import re

SYNTAX_SEPARATOR_REGEX = re.compile(r"(?:[^/]|\\/)+")

exceptions = (
    IndexError,
    KeyError,
    TypeError,
    ValueError,
)


def quickson_navigator(data, *quicksons):

    while quicksons:

        if not isinstance(data, (list, dict)):
            raise TypeError("Data must be a dict or list")

        worker, *rest = quicksons

        if isinstance(data, list):
            if not worker.isdigit():
                raise ValueError("List index must be an integer")

            data = data[int(worker)]
        else:
            data = data[worker]

        quicksons = rest

    return data


def navigate_quickson(quickson_string, data, default=None):
    """
    A function that navigates through a quickson string and returns the data from
    the data dictionary.
    """

    try:
        return quickson_navigator(
            data, *SYNTAX_SEPARATOR_REGEX.findall(quickson_string)
        )
    except exceptions as _:
        if default is None:
            raise _

    return default


if __name__ == "__main__":
    # fmt: off
    assert navigate_quickson("a/b/c", {"a": {"b": {"c": "d"}}}) == "d"
    assert navigate_quickson("a/b/c/0", {"a": {"b": {"c": ["d"]}}}) == "d"
    assert navigate_quickson("a/b/c/0/1", {"a": {"b": {"c": [{"d": []}]}}}, "test") == "test"
    # fmt: on



from setuptools import find_packages, setup

from anchor import __version__

setup(
    name="anchor-kr",
    version=__version__,
    author="kr@justfoolingaround",
    author_email="kr.justfoolingaround@gmail.com",
    description="Me, you and anchor make your scraper complete.",
    packages=find_packages(),
    url="https://github.com/justfoolingaround/anchor",
)
